/**
 * Copyright (c) 2006-2012 EBM WebSourcing, 2012-2016 Linagora
 * 
 * This program/library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or (at your
 * option) any later version.
 * 
 * This program/library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program/library; If not, see http://www.gnu.org/licenses/
 * for the GNU Lesser General Public License version 2.1.
 */
package org.objectweb.petals.tools.rmi.server.remote.implementations;

import java.io.IOException;
import java.rmi.RemoteException;
import java.rmi.server.UnicastRemoteObject;
import java.util.logging.Logger;

import javax.jbi.JBIException;
import javax.jbi.management.MBeanNames;
import javax.jbi.servicedesc.ServiceEndpoint;
import javax.management.MBeanServer;
import javax.naming.InitialContext;
import javax.xml.namespace.QName;
import javax.xml.parsers.ParserConfigurationException;
import javax.xml.transform.TransformerException;

import org.objectweb.petals.tools.rmi.remote.RemoteComponentContext;
import org.objectweb.petals.tools.rmi.serializable.MBeanNamesImpl;
import org.objectweb.petals.tools.rmi.server.remote.implementations.RemoteComponentContextClientFactoryImpl.SharedComponentData;
import org.objectweb.petals.tools.rmi.server.util.Convert;
import org.ow2.easywsdl.wsdl.WSDLFactory;
import org.ow2.easywsdl.wsdl.api.Description;
import org.ow2.easywsdl.wsdl.api.Endpoint;
import org.ow2.easywsdl.wsdl.api.InterfaceType;
import org.ow2.easywsdl.wsdl.api.Service;
import org.ow2.easywsdl.wsdl.api.WSDLException;
import org.ow2.easywsdl.wsdl.api.WSDLWriter;
import org.ow2.easywsdl.wsdl.api.abstractItf.AbsItfDescription.WSDLVersionConstants;
import org.ow2.petals.commons.log.Level;
import org.ow2.petals.component.framework.logger.ConsumeExtFlowStepEndLogData;
import org.ow2.petals.component.framework.logger.ConsumeExtFlowStepFailureLogData;
import org.ow2.petals.component.framework.util.ServiceEndpointKey;
import org.w3c.dom.Document;
import org.w3c.dom.DocumentFragment;
import org.xml.sax.SAXException;

/**
 * The RMI implementation of component context.
 * @author Nicolas Salatge - EBM WebSourcing
 */
public class RemoteComponentContextImpl extends UnicastRemoteObject implements RemoteComponentContext {

    /**
     * the serialVersionUID.
     */
    private static final long serialVersionUID = 7288972323223181501L;

    /**
     * the MBean Names.
     */
    private final MBeanNames mbeanNames;

    private final Logger logger;

    /**
     * The WSDL Factory lazy singleton.
     */
    private static WSDLFactory wsdlFactory;

    /**
     * The WSDL Writer lazy singleton.
     */
    private static WSDLWriter wsdlWriter;

    private final SharedComponentData sharedData;

    private final String rmiClientId;

    public RemoteComponentContextImpl(final SharedComponentData sharedData, final String rmiClientId)
            throws JBIException, RemoteException {
        this.sharedData = sharedData;
        this.rmiClientId = rmiClientId;
        this.mbeanNames = new MBeanNamesImpl(this.sharedData.componentContext.getMBeanNames().getJmxDomainName());
        this.logger = this.sharedData.componentContext.getLogger("rmi-server", null);
    }

    @Override
    public ServiceEndpoint activateEndpoint(final QName interfaceName, final QName serviceName,
            final String endpointName) throws JBIException, RemoteException {

        final ServiceEndpointKey key = new ServiceEndpointKey(serviceName, endpointName);

        final Document wsdl;
        try {
            wsdl = generateWSDL(interfaceName, serviceName, endpointName);
        } catch (final WSDLException | SAXException | IOException | ParserConfigurationException
                | TransformerException e) {
            RemoteException exception = new RemoteException("Error occured during the WSDL generation");
            exception.setStackTrace(e.getStackTrace());
            throw exception;
        }

        try {
            this.sharedData.ownedEndpointDescriptor.put(key, wsdl);
            this.sharedData.ownedEndpointProviders.put(key, this.rmiClientId);
            
            final ServiceEndpoint se = this.sharedData.componentContext.activateEndpoint(serviceName, endpointName);

            this.logger.fine("New Service Endpoint deployed: " + se);
            
            return Convert.convertConcreteServiceEndpointToFakeServiceEndpoint(se);
            
        } catch (final JBIException e) {
            this.sharedData.ownedEndpointDescriptor.remove(key);
            this.sharedData.ownedEndpointProviders.remove(key);

            throw Convert.convertJBIException(e);
        }
    }

    @Override
    public void deactivateEndpoint(final ServiceEndpoint endpoint) throws JBIException,
            RemoteException {

        final ServiceEndpointKey key = new ServiceEndpointKey(endpoint);
        try {
            final ServiceEndpoint se = this.sharedData.componentContext.getEndpoint(endpoint.getServiceName(),
                    endpoint.getEndpointName());
            this.sharedData.componentContext.deactivateEndpoint(se);
        } catch (JBIException e) {
            throw Convert.convertJBIException(e);
        } finally {
            this.sharedData.ownedEndpointDescriptor.remove(key);
            this.sharedData.ownedEndpointProviders.remove(key);
        }
    }

    @Override
    public void deregisterExternalEndpoint(final ServiceEndpoint externalEndpoint)
            throws JBIException, RemoteException {
        throw new UnsupportedOperationException("deregisterExternalEndpoint: Not implemented.");
    }

    @Override
    public String getComponentName() throws RemoteException {
        return this.sharedData.componentContext.getComponentName();
    }

    @Override
    public ServiceEndpoint getEndpoint(final QName service, final String name)
            throws RemoteException {
        final ServiceEndpoint seC = this.sharedData.componentContext.getEndpoint(service, name);
        return seC != null ? Convert.convertConcreteServiceEndpointToFakeServiceEndpoint(seC) : null;
    }

    @Override
    public Document getEndpointDescriptor(final ServiceEndpoint endpoint) throws JBIException,
            RemoteException {
        // we need to use the concrete SE
        final ServiceEndpoint se = this.sharedData.componentContext.getEndpoint(endpoint.getServiceName(),
                endpoint.getEndpointName());
        try {
            return se != null ? this.sharedData.componentContext.getEndpointDescriptor(se) : null;
        } catch (final JBIException e) {
            throw Convert.convertJBIException(e);
        }
    }

    private static ServiceEndpoint[] convertSEArray(final ServiceEndpoint[] sesTemp) {
        final ServiceEndpoint[] sesRes;
        sesRes = new ServiceEndpoint[sesTemp.length];
        for (int i = 0; i < sesRes.length; i++) {
            sesRes[i] = Convert.convertConcreteServiceEndpointToFakeServiceEndpoint(sesTemp[i]);
        }
        return sesRes;
    }

    @Override
    public ServiceEndpoint[] getEndpoints(final QName interfaceName) throws RemoteException {
        final ServiceEndpoint[] sesTemp = this.sharedData.componentContext.getEndpoints(interfaceName);
        return sesTemp != null ? convertSEArray(sesTemp) : null;
    }

    @Override
    public ServiceEndpoint[] getEndpointsForService(final QName serviceName) throws RemoteException {
        final ServiceEndpoint[] sesTemp = this.sharedData.componentContext.getEndpointsForService(serviceName);
        return sesTemp != null ? convertSEArray(sesTemp) : null;
    }

    @Override
    public ServiceEndpoint[] getExternalEndpoints(final QName interfaceName) throws RemoteException {
        final ServiceEndpoint[] sesTemp = this.sharedData.componentContext.getExternalEndpoints(interfaceName);
        return sesTemp != null ? convertSEArray(sesTemp) : null;
    }

    @Override
    public ServiceEndpoint[] getExternalEndpointsForService(final QName serviceName)
            throws RemoteException {
        final ServiceEndpoint[] sesTemp = this.sharedData.componentContext.getExternalEndpointsForService(serviceName);
        return sesTemp != null ? convertSEArray(sesTemp) : null;
    }

    @Override
    public String getInstallRoot() throws RemoteException {
        return this.sharedData.componentContext.getInstallRoot();
    }

    @Override
    public Logger getLogger(final String suffix, final String resourceBundleName)
            throws JBIException, RemoteException {
        // A logger is not serializable
        throw new UnsupportedOperationException("getLogger: Not implemented.");
    }

    @Override
    public void logConsumesExtFlowStepBegin(final String flowInstanceId, final String flowStepId,
            final String flowInterfaceName, final String flowServiceName, final String flowEndpointName,
            final String flowOperationName) throws RemoteException {
        log(Level.MONIT, "", new RMIConsumeExtFlowStepBeginLogData(flowInstanceId, flowStepId, flowInterfaceName,
                flowServiceName, flowEndpointName, flowOperationName));
    }

    @Override
    public void logConsumesExtFlowStepEnd(final String flowInstanceId, final String flowStepId) throws RemoteException {
        log(Level.MONIT, "", new ConsumeExtFlowStepEndLogData(flowInstanceId, flowStepId));
    }

    @Override
    public void logConsumesExtFlowStepFailure(String flowInstanceId, String flowStepId, final String msg)
            throws RemoteException {
        log(Level.MONIT, "", new ConsumeExtFlowStepFailureLogData(flowInstanceId, flowStepId, msg));
    }

    @Override
    public void log(final java.util.logging.Level level, final String msg, final Object param1) throws RemoteException {
        this.logger.log(level, msg, param1);
    }

    @Override
    public MBeanNames getMBeanNames() throws RemoteException {
        return this.mbeanNames;
    }

    @Override
    public MBeanServer getMBeanServer() throws RemoteException {
        throw new NoSuchMethodError("getMBeanServer: Not implemented.");
    }

    @Override
    public InitialContext getNamingContext() throws RemoteException {
        throw new NoSuchMethodError("getNamingContext: Not implemented.");
    }

    @Override
    public Object getTransactionManager() throws RemoteException {
        throw new NoSuchMethodError("getTransactionManager: Not implemented.");
    }

    @Override
    public String getWorkspaceRoot() throws RemoteException {
        return this.sharedData.componentContext.getWorkspaceRoot();
    }

    @Override
    public void registerExternalEndpoint(final ServiceEndpoint externalEndpoint)
            throws JBIException, RemoteException {
        throw new NoSuchMethodError("registerExternalEndpoint: Not implemented.");
    }

    @Override
    public ServiceEndpoint resolveEndpointReference(final DocumentFragment epr)
            throws RemoteException {
        final ServiceEndpoint seC = this.sharedData.componentContext.resolveEndpointReference(epr);
        return seC != null ? Convert.convertConcreteServiceEndpointToFakeServiceEndpoint(seC) : null;
    }

    /**
     * Allow to generate a WSDL file by providing, interface/service/endpoint name
     */
    private Document generateWSDL(final QName interfaceName, final QName serviceName,
            final String endpointName) throws WSDLException, SAXException, IOException,
            ParserConfigurationException, TransformerException {
        final Description description = getWSDLFactory()
                .newDescription(WSDLVersionConstants.WSDL20);

        // Add service/interface/endpoint to the description
        description.setTargetNamespace(serviceName.getNamespaceURI());
        final Service service = description.createService();
        service.setQName(serviceName);
        final InterfaceType _interface = description.createInterface();
        _interface.setQName(interfaceName);
        final Endpoint endpoint = service.createEndpoint();
        endpoint.setName(endpointName);
        service.addEndpoint(endpoint);
        service.setInterface(_interface);
        description.addService(service);
        description.addInterface(_interface);

        return getWSDLWriter().getDocument(description);
    }

    /**
     * Lazy static instantiation.
     */
    private static final synchronized WSDLFactory getWSDLFactory() throws WSDLException {
        if (wsdlFactory == null) {
            wsdlFactory = WSDLFactory.newInstance();
        }

        return wsdlFactory;
    }

    /**
     * Lazy static instantiation.
     */
    private static final synchronized WSDLWriter getWSDLWriter() throws WSDLException {
        if (wsdlWriter == null) {
            wsdlWriter = getWSDLFactory().newWSDLWriter();
        }

        return wsdlWriter;
    }
}
